//==================================================
//Auto pilot(prototype2)
//
//MPU board: mbed LPC1768
//Multiplexer TC74HC157AP
//Accelerometer +Gyro sensor : BMX055
//Pressure sensor BME280
//GPS module AE-GPS
//
//2020/3/11 A.Toda
//==================================================

#include "mbed.h"

#include "stdio.h"
#include "math.h"
#include "errno.h"
#include "ctype.h"

//Header file for pressure sensor
#include "BME280.h"

//Header file for inertial sensor
#include "test_BMX055.h"

//==================================================

#define INITIAL_ALTITUDE 0.0
#define DIGIT_TO_VAL(_x)        (_x - '0')
#define RAD_TO_DEG 57.3

#define MAG_OFFSET_X 0.0
#define MAG_OFFSET_Y 0.0
#define MAG_OFFSET_Z 0.0
//==================================================
//Port Setting
DigitalOut myled(LED1);

BME280 pressure_sensor(p9, p10);//(sda, scl);
//test_BMX055 inertial_sensor(p28, p27);//(sda, scl);
test_BMX055 inertial_sensor(p9, p10);//(sda, scl);

Serial pc(USBTX, USBRX); // tx, rx
Serial gps(p13,p14);//tx,rx

DigitalIn gps_sw(p11);

//=========================================================
//Ticker
Ticker timer1; //for gps notification
Ticker timer2; //for attitude calculation

//Timer
Timer passed_time;
int dummy_counter;
//=================================================
//Ticker variables
//=================================================
int gps_notification_rate = 25; //usec
int inertial_sensor_rate = 200; //Hz

//==================================================
//Timer variables
//==================================================
float time_new,time_old;
float feedback_rate = 10000; //usec

//==================================================
//Vetor variables
float quaternion[4][1],pre_quaternion[4][1],damp_1[4][1];//クォータニオン

//==================================================
//GPS variables
//==================================================
int fp_count=0;
int h_time=0,m_time=0,s_time=0;

int notification_from_sub;

float g_hokui,g_tokei;
float gps_azi;
float speed_m;

//==================================================
//Pressure variables
//==================================================
float pressure_1,temperature_1;
float altitude_1,altitude_offset;

//==================================================
//Inertial sonsor  variables
//==================================================
float gx,gy,gz;
float ax,ay,az;

float gx_mean,gy_mean,gz_mean;
int mean_counter=2000;

float limit_acc;

float euler_roll_pitch[2];

//==================================================
//Geomagnetrometer  variables
//==================================================
float mx,my,mz;
float yaw;

//==================================================
//Predeclaration for functions
//==================================================

void gps_receiver();
void gps_receiver_2();
void gps_receiver_3();

void gps_notification();
void inertial_sensor_function();

float get_altitude(float offset);

//=========================================================
//Gps receiver function
//=========================================================
void gps_receiver()
{

    //variables
    int numbers_of_nmea;
    char gps_data[256];



    numbers_of_nmea=0;
    while(gps.getc()!='$') {
    }

    while( (gps_data[numbers_of_nmea]=gps.getc()) != '\r') {

        numbers_of_nmea++;
        if(numbers_of_nmea==256) {
            numbers_of_nmea=255;
            break;
        }
    }

    gps_data[numbers_of_nmea]='\0';

    //char gps_datae[256]="$GNRMC,222219.000,A,3605.4010,N,14006.8240,E,0.11,109.92,191016,,,A";
    if( sscanf(gps_data, "GNRMC,%f,%f,%f,%f\r",&g_hokui,&g_tokei,&gps_azi,&speed_m) >= 1) {
        //pc.printf("変換前データ %f,%f,%f\r\n",g_hokui,g_tokei,azi);
    } else {
    }//if ends
}

//=========================================================
//Gps receiver function2
//=========================================================
void gps_receiver_2()
{

    //variables
    int numbers_of_nmea;
    char gps_data[256];
    char lat_data[10];
    char long_data[11];
    char azi_data[11];

    char *ep;

    numbers_of_nmea=0;

    int commma_counter=0;

    while(gps.getc()!='$') {
    }

    while( (gps_data[numbers_of_nmea]=gps.getc()) != '\r') {

        numbers_of_nmea++;
        if(numbers_of_nmea==256) {
            numbers_of_nmea=255;
            break;
        }
    }

    gps_data[numbers_of_nmea]='\0';

    for(int nmea_counter=0; nmea_counter<34; nmea_counter++) {

        if(gps_data[nmea_counter]==',') {

            if(commma_counter==0) {
                for(int lat_counter=0; lat_counter<9; lat_counter++) {
                    lat_data[lat_counter]=gps_data[nmea_counter+lat_counter+1];
                }
                lat_data[9]='\0';
                g_hokui = float(strtod( lat_data, &ep ));

                commma_counter++;

            } else if(commma_counter==1) {
                for(int long_counter=0; long_counter<10; long_counter++) {
                    //pc.putc(gps_data[nmea_counter+long_counter+1]);
                    long_data[long_counter]=gps_data[nmea_counter+long_counter+1];
                }
                long_data[10]='\0';
                g_tokei = float(strtod( long_data, &ep ));

                commma_counter++;

            } else if(commma_counter==2) {
                for(int azi_counter=0; azi_counter<5; azi_counter++) {
                    azi_data[azi_counter]=gps_data[nmea_counter+azi_counter+1];
                }
                azi_data[5]='\0';
                gps_azi = float(strtod( azi_data, &ep ));

            } else {}
        } else {}
    }
}

//=========================================================
//Gps receiver function3
//=========================================================
void gps_receiver_3()
{

    //variables
    int numbers_of_nmea;

    int i_hokui,i_tokei,i_azi;

    char gps_data[256];

    numbers_of_nmea=0;
    i_hokui=0;
    i_tokei=0;
    i_azi=0;

    int commma_counter;
    commma_counter=0;

    while(gps.getc()!='$') {
    }

    while( (gps_data[numbers_of_nmea]=gps.getc()) != '\r') {

        numbers_of_nmea++;
        if(numbers_of_nmea==256) {
            numbers_of_nmea=255;
            break;
        }
    }

    gps_data[numbers_of_nmea]='\0';

    for(int nmea_counter=0; nmea_counter<34; nmea_counter++) {

        if(gps_data[nmea_counter]==',') {
            commma_counter++;
        } else {}

        //pc.printf("%d\r\n",commma_counter);

        if(commma_counter==1) {


            if(isdigit(gps_data[nmea_counter])) {

                i_hokui+=DIGIT_TO_VAL(gps_data[nmea_counter]);
                i_hokui*=10;
            } else {}

        } else if(commma_counter==2) {

            if(isdigit(gps_data[nmea_counter])) {

                i_tokei+=DIGIT_TO_VAL(gps_data[nmea_counter]);
                i_tokei*=10;
                //pc.printf("%d\r\n",i_tokei);
            } else {}

        } else if(commma_counter==3) {
            if(isdigit(gps_data[nmea_counter])) {

                i_azi+=DIGIT_TO_VAL(gps_data[nmea_counter]);
                i_azi*=10;
            } else {}

        } else {}

    }

    g_hokui = float(i_hokui)/10000000;
    g_tokei=float(i_tokei)/10000000;
    gps_azi = float(i_azi)/10.0;

}

//=========================================================
//gps notification function
//=========================================================
void gps_notification()
{
    if(gps_sw==1) {
        notification_from_sub=1;
    } else {
        notification_from_sub=0;
    }
}

//=========================================================
//inertial sensor acquisition function
//=========================================================
//atitude calculation from acceleromter
void atitude_estimation_from_accelerometer(float ax,float ay,float az,float *roll_and_pitch){
    
    roll_and_pitch[0] = atan(ay/az)*RAD_TO_DEG;//roll
    roll_and_pitch[1] = atan(ax/sqrt( (ay*ay+az*az) ) )*RAD_TO_DEG;//pitch
    
    return;
}

void euler_to_quaternion(float *qua,float *roll_pitch,float yaw_temp){
    
    float Roll = roll_pitch[0]/RAD_TO_DEG;
    float Pitch = roll_pitch[1]/RAD_TO_DEG;
    float Yaw = yaw_temp/RAD_TO_DEG;
    
    qua[0]=cos(Roll/2.0)*cos(Pitch/2.0)*cos(Yaw/2.0)+sin(Roll/2.0)*sin(Pitch/2.0)*sin(Yaw/2.0);
    qua[1]=sin(Roll/2.0)*cos(Pitch/2.0)*cos(Yaw/2.0)-cos(Roll/2.0)*sin(Pitch/2.0)*sin(Yaw/2.0);
    qua[2]=cos(Roll/2.0)*sin(Pitch/2.0)*cos(Yaw/2.0)+sin(Roll/2.0)*cos(Pitch/2.0)*sin(Yaw/2.0);
    qua[3]=cos(Roll/2.0)*cos(Pitch/2.0)*sin(Yaw/2.0)-sin(Roll/2.0)*sin(Pitch/2.0)*cos(Yaw/2.0);
    
}

void Quaternion_to_Euler(float *qua,float *roll_pitch_yaw){
    
    //float pitch,roll,yaw;
    float q0=qua[0];
    float q1=qua[1];
    float q2=qua[2];
    float q3=qua[3];
    
    roll_pitch_yaw[0]=-RAD_TO_DEG*atan2(( 2*(q2*q3+q0*q1)), (q0*q0-q1*q1-q2*q2+q3*q3) );
    roll_pitch_yaw[1]=-RAD_TO_DEG*asin((2*(q1*q3-q0*q2)));
    roll_pitch_yaw[2] = float ( atan2(2*(q1*q2+q0*q3),(q0*q0+q1*q1-q2*q2-q3*q3)) )*RAD_TO_DEG;

}

float tilt_compensated_eCompass(float mag_x,float mag_y,float mag_z,float *roll_pitch){
    
    float yaw_temp;
    float roll_rad=roll_pitch[0]/RAD_TO_DEG;
    float pitch_rad=roll_pitch[1]/RAD_TO_DEG;
    
    float B_y=(mag_z-MAG_OFFSET_Z)*sin(roll_rad)-(mag_y-MAG_OFFSET_Y)*cos(roll_rad);
    float B_x=(mag_x-MAG_OFFSET_X)*cos(pitch_rad)+(mag_y-MAG_OFFSET_Y)*sin(pitch_rad)*sin(roll_rad)
                                            +(mag_z-MAG_OFFSET_Z)*sin(pitch_rad)*cos(roll_rad);
    
    yaw_temp = float (atan2(double(-B_y),double(B_x)) )*RAD_TO_DEG;//度数表現(-180to180)
        
    //pc.printf("mag_x=%f,mag_y=%f,mag_z=%f,B_y=%f,B_x=%f,yaw=%f\r\n",mag_x,mag_y,mag_z,B_y,B_x,yaw_temp);
                               
    
    return yaw_temp;
}



void inertial_sensor_function()
{

    //detach gps notification polling
    timer1.detach();
    
    int counter;
    //dummy_counter++;

    gx=inertial_sensor.get_gyro_x_data();
    gy=inertial_sensor.get_gyro_y_data();
    gz=inertial_sensor.get_gyro_z_data();

    ax = inertial_sensor.get_acc_x_data();
    ay = inertial_sensor.get_acc_y_data();
    az = inertial_sensor.get_acc_z_data();

    mx = inertial_sensor.get_mag_x_data();
    my = inertial_sensor.get_mag_y_data();
    mz = inertial_sensor.get_mag_z_data();
    
    gx -=gx_mean; 
    gy -=gy_mean;
    gz -=gz_mean;
    
    //pc.printf("%f,%f,%f,%f,%f,%f\r\n",gx,gy,gz,ax,ay,az);

    float absolute_acc = sqrt(ax*ax+ay*ay+az*az);
    
    //最新の時間の取得
    time_new=passed_time.read();
    
    //------------------------------------
    //クォータニオンの計算
    float dt_1 = time_new-time_old;
    float dt_2 = dt_1*0.5;
    
    //pc.printf("%f,",dt_1);
    //pc.printf("%f,%f,%f,%f,%f,%f\r\n",gx,gy,gz,ax,ay,az);
    
    float q0_n,q1_n,q2_n,q3_n;
    float q0_o,q1_o,q2_o,q3_o;
    
    q0_o=pre_quaternion[0][0];
    q1_o=pre_quaternion[1][0];
    q2_o=pre_quaternion[2][0];    
    q3_o=pre_quaternion[3][0];
    
    //クォータニオン積分(更新) 
    q0_n = q0_o+dt_2*(-q1_o*gx-q2_o*gy-q3_o*gz);
    q1_n = q1_o+dt_2*(q0_o*gx-q3_o*gy+q2_o*gz);
    q2_n = q2_o+dt_2*(q3_o*gx+q0_o*gy-q1_o*gz);
    q3_n = q3_o+dt_2*(-q2_o*gx+q1_o*gy+q0_o*gz);    
    
    //正規化
    float sum_qua=0.0;
    float q0q0,q1q1,q2q2,q3q3;
    q0q0=q0_n*q0_n;
    q1q1=q1_n*q1_n;
    q2q2=q2_n*q2_n;
    q3q3=q3_n*q3_n;
    
    sum_qua=float(sqrt(double(q0q0+q1q1+q2q2+q3q3)));
    q0_n/=sum_qua;
    q1_n/=sum_qua;
    q2_n/=sum_qua;
    q3_n/=sum_qua;
    
    //pc.printf("%f,%f,%f,%f\r\n",q0_n,q1_n,q2_n,q3_n);
    //pc.printf("%f\r\n",dt_1);
    
    if((absolute_acc>0.09*limit_acc)&&(absolute_acc<1.01*limit_acc)){
        
        //加速度による補正
        float q0_a,q1_a,q2_a,q3_a;
        float acc_euler_roll_pitch[2]={};
        float acc_quaternion[4][1]={};
        
        float yaw_mag;
        
        atitude_estimation_from_accelerometer(ax,ay,az,acc_euler_roll_pitch);
        //------------------------------------------
        //roll,pitch角を使用した地磁気センサの補正とyaw角の計算
        yaw_mag = tilt_compensated_eCompass(mx,my,mz,acc_euler_roll_pitch);
        //pc.printf("Yaw_mag=%f",yaw_mag);
        
        euler_to_quaternion(acc_quaternion[0],acc_euler_roll_pitch,yaw_mag);
        
        q0_a=acc_quaternion[0][0];
        q1_a=acc_quaternion[1][0];
        q2_a=acc_quaternion[2][0];
        q3_a=acc_quaternion[3][0];
        
        //acc_euler_roll_pitch
        //pc.printf("%f,%f\r\n",acc_euler_roll_pitch[0],acc_euler_roll_pitch[1]);
        //pc.printf("%f,%f,%f,%f\r\n",q0_a,q1_a,q2_a,q3_a);
        
        q0_n = q0_n*0.98+q0_a*0.05;
        q1_n = q1_n*0.98+q1_a*0.05;
        q2_n = q2_n*0.98+q2_a*0.05;
        q3_n = q3_n*0.98+q3_a*0.05;
        
        }else{}
    
    quaternion[0][0]=q0_n;
    quaternion[1][0]=q1_n;
    quaternion[2][0]=q2_n;
    quaternion[3][0]=q3_n;

    //離散時間の更新
    time_old=time_new;
    
    //euler_to_quaternion(quaternion[0],euler_roll_pitch);
    float euler_roll_pitch_yaw[3]={};
    Quaternion_to_Euler(quaternion[0],euler_roll_pitch_yaw);
    
    euler_roll_pitch[0]=euler_roll_pitch_yaw[0];
    euler_roll_pitch[1]=euler_roll_pitch_yaw[1];
    yaw=euler_roll_pitch_yaw[2];
    
    //pc.printf("Yaw=%f\r\n",yaw);
    
    for(counter =0;counter<4;counter++){
        pre_quaternion[counter][0] = quaternion[counter][0];
        }
        
    //timer1: gps notification polling, 40 kHz
    timer1.attach_us(&gps_notification, gps_notification_rate);

}

//=========================================================
//altitude function
//=========================================================
float get_altitude(float offset)
{

    pressure_1=pressure_sensor.getPressure();
    temperature_1=pressure_sensor.getTemperature();

    float altitude=pow((1013.25/pressure_1),0.190);
    altitude-=1.0;
    altitude*=(temperature_1+273.15);
    altitude/=0.0065;
    altitude-=offset;
    altitude+=INITIAL_ALTITUDE;

    return altitude;

}

//=========================================================
//Timer function
//=========================================================
void timer_reset()
{
    passed_time.reset();//reset timer counter
}

void timer_read()
{
    time_new=passed_time.read();
    //pc.printf("経過時間=%f\r\n",time_new);
}

//=========================================================
// main
//=========================================================
int main()
{
    //UART initialization
    pc.baud(115200); //115.2 kbps
    gps.baud(115200); //115.2 kbps

    //Range setting for inertial sensor
    //慣性センサのレンジ設定
    inertial_sensor.set_gyro_range(1);
    inertial_sensor.set_acc_range(2);

    //地磁気センサの初期化
    inertial_sensor.initiate_mag();
    
    //Initialization for pressure sensor
    //気圧センサの初期化
    pressure_sensor.initialize();

    notification_from_sub=0;//サブマイコンからの通知を初期化
    dummy_counter=0;

    //-------------------------------------------
    //altitude initialization
    /*
    2回目からの計測でまともな高度が得られるので、オフセット計測を2回する
    */
    //-------------------------------------------
    altitude_offset=get_altitude(0.0);//初期高度を取得（オフセットの取得)
    wait_us (100000);//wait 0.1 s
    altitude_offset=get_altitude(0.0);//初期高度を取得（オフセットの取得）


    //-------------------------------------------
    //初期クォータニオンの取得
    ax = inertial_sensor.get_acc_x_data();
    ay = inertial_sensor.get_acc_y_data();
    az = inertial_sensor.get_acc_z_data();

    mx = inertial_sensor.get_mag_x_data();
    my = inertial_sensor.get_mag_y_data();
    mz = inertial_sensor.get_mag_z_data();
    
    limit_acc = sqrt(ax*ax+ay*ay+az*az);
    
    atitude_estimation_from_accelerometer(ax,ay,az,euler_roll_pitch);

    //------------------------------------------
    //roll,pitch角を使用した地磁気センサの補正とyaw角の計算
    yaw = tilt_compensated_eCompass(mx,my,mz,euler_roll_pitch);
    
    euler_to_quaternion(pre_quaternion[0],euler_roll_pitch,yaw);
    
    //------------------------------------------
    //ジャイロオフセットの計算
    gx_mean=gy_mean=gz_mean=0.0;
    
    for(int counter=0;counter<mean_counter;counter++){
        
        gx=inertial_sensor.get_gyro_x_data();
        gy=inertial_sensor.get_gyro_y_data();
        gz=inertial_sensor.get_gyro_z_data();
    
        gx_mean+=gx;
        gy_mean+=gy;
        gz_mean+=gz;
        wait_us(10000);
        }
    
    gx_mean/=mean_counter;
    gy_mean/=mean_counter;
    gz_mean/=mean_counter;

    //-------------------------------------------
    //Ticker
    //-------------------------------------------
    //timer1: gps notification polling, 40 kHz
    timer1.attach_us(&gps_notification, gps_notification_rate);
    //timer2: acquisition, 200 Hz
    timer2.attach(&inertial_sensor_function, 1.0f/inertial_sensor_rate);

    //-------------------------------------------
    //Timer starts
    //-------------------------------------------
    passed_time.start();
    
    pc.printf("Activated.\r\n");
    while(1) {
        if(notification_from_sub==1) {
            
            timer2.detach();
            timer1.detach();
                    
            myled=1;

            //gpsプロセッサからのシリアル信号を受信
            //gps_receiver_3();
            gps_receiver();
            notification_from_sub=0;

            altitude_1=get_altitude(altitude_offset);

            //デバッグの為の表示
            pc.printf("%f,%f,%f,%f,%f\r\n",g_hokui,g_tokei,altitude_1,speed_m,gps_azi);
            //pc.printf("%f,%f\r\n",g_hokui,g_tokei);
        
        } else {
            myled=0;
            timer2.detach();
        }
        
        pc.printf("%f,%f\r\n",euler_roll_pitch[0],euler_roll_pitch[1]);
        //pc.printf("%f,%f,%f\r\n",euler_roll_pitch[0],euler_roll_pitch[1],yaw);
        //pc.printf("%f,%f,%f,%f\r\n",quaternion[0][0],quaternion[1][0],quaternion[2][0],quaternion[3][0]);
        
        timer1.attach_us(&gps_notification, gps_notification_rate);
        timer2.attach(&inertial_sensor_function,1.0f/inertial_sensor_rate);
        
        
        wait_us(10000);//waits for 10ms
    }

}//int main ends